import pandas as pd
from sklearn.preprocessing import StandardScaler, LabelEncoder
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
import tensorflow as tf
from tensorflow.keras.layers import Dense, Dropout, BatchNormalization, Add, Activation, Input
from tensorflow.keras.callbacks import Callback
import joblib
import os
import queue
import threading
import numpy as np
from tensorflow.keras.regularizers import l2
import random
from itertools import product
import chardet

class ClassificationDL:
    def __init__(self, dataset_url, architecture, hyperparameters, model_name, target_col):
        physical_devices = tf.config.list_physical_devices('GPU')
        if physical_devices:
            print("GPU(s) detected:", len(physical_devices))
            for device in physical_devices:
                try:
                    tf.config.experimental.set_memory_growth(device, True)
                    print(f"Memory growth enabled for GPU device: {device}")
                    
                except RuntimeError as e:
                    print(f"Error configuring GPU device {device}: {e}")
            
            tf.keras.mixed_precision.set_global_policy('mixed_float16')
            print("Mixed precision policy enabled for GPU training")
        else:
            print("No GPU devices detected. Running on CPU.")
            
        self.dataset_url = dataset_url
        self.architecture = architecture
        self.hyperparameters = hyperparameters
        self.model_name = model_name
        self.model_path = f'{model_name}.h5'
        self.scaler_path = f'scaler_{model_name}.pkl'
        self.label_encoder_path = f'label_encoder_{model_name}.pkl'
        self.directory_path = "models"
        self.complete_model_path = os.path.join(self.directory_path, self.model_path)
        self.complete_scaler_path = os.path.join(self.directory_path, self.scaler_path)
        self.complete_label_encoder_path = os.path.join(self.directory_path, self.label_encoder_path)
        self.target_col = target_col
        
        # Initialize these variables
        self.current_epoch_info = None
        self.epoch_info_queue = None
        
        self.load_data()
        self.preprocess_data()

    def load_data(self):
        with open(self.dataset_url, 'rb') as file:
            result = chardet.detect(file.read())
            self.encoding = result['encoding']
        self.df = pd.read_csv(self.dataset_url, encoding=self.encoding)        
        self.df = self.df.dropna()
        
        def contains_url(column):
            if column.dtype == 'object':
                return column.str.contains(r'http[s]?://', na=False).any()
            return False

        url_columns = [col for col in self.df.columns if contains_url(self.df[col])]
        self.df = self.df.drop(columns=url_columns)
        
        target_idx = self.df.columns.get_loc(self.target_col)
        self.X = self.df.drop(columns=[self.target_col]) 
        self.y = self.df.iloc[:, target_idx]

        self.label_encoders = {}
        for column in self.X.select_dtypes(include=['object']).columns:
            le = LabelEncoder()
            self.X[column] = le.fit_transform(self.X[column])
            self.label_encoders[column] = le

        if self.y.dtype == object:
            le = LabelEncoder()
            self.y = le.fit_transform(self.y)
            self.label_encoders['target'] = le

        # Store column names for inference
        self.feature_columns = self.X.columns.tolist()
        
        self.unique_labels = np.unique(self.y)
        self.label_map = {label: i for i, label in enumerate(self.unique_labels)}
        self.y = np.array([self.label_map[label] for label in self.y])
        for column, le in self.label_encoders.items():
            print(f"Column: {column}, Classes: {list(le.classes_)}")

    def preprocess_data(self):
        self.scaler = StandardScaler()
        self.X_standardized = self.scaler.fit_transform(self.X)
        self.X = self.X_standardized
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            self.X, self.y, test_size=0.2, random_state=42)

    def create_model(self, batch_size, epochs, dropout_rate, learning_rate, dense_units):
        unique_values = np.unique(self.y)
        self.is_binary = len(unique_values) <= 2
        num_classes = len(unique_values)

        inputs = Input(shape=(self.X_train.shape[1],))
        x = inputs

        for i, units in enumerate(dense_units):
            x = Dense(units, kernel_regularizer=l2(0.001), activation="relu")(x)
            x = Dropout(dropout_rate)(x)
            
            if i % 2 != 0:
                x = BatchNormalization()(x)

        if self.is_binary:
            outputs = Dense(1, activation='sigmoid')(x)
            loss = 'binary_crossentropy'
        else:
            outputs = Dense(num_classes, activation='softmax')(x)
            loss = 'sparse_categorical_crossentropy'

        self.model = tf.keras.Model(inputs=inputs, outputs=outputs)
        self.model.compile(
            optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate),
            loss=loss,
            metrics=['accuracy']
        )

    def train_model(self, batch_size, epochs, dropout_rate, learning_rate, dense_units, testing=False):
        self.create_model(batch_size, epochs, dropout_rate, learning_rate, dense_units)

        self.epoch_data = []
        self.current_val_acc = 0
        self.epoch_info_queue = queue.Queue()

        class CustomCallback(Callback):
            def __init__(self, outer_instance):
                super().__init__()
                self.outer_instance = outer_instance

            def on_epoch_end(self, epoch, logs=None):
                test_loss, test_acc = self.model.evaluate(
                    self.outer_instance.X_test, 
                    self.outer_instance.y_test, 
                    verbose=0
                )
                train_loss, train_acc = logs['loss'], logs['accuracy']

                epoch_info = {
                    "epoch": epoch + 1,
                    "train_loss": train_loss,
                    "train_acc": train_acc,
                    "test_loss": test_loss,
                    "test_acc": test_acc
                }
                self.outer_instance.epoch_data.append(epoch_info)
                self.outer_instance.current_epoch_info = epoch_info
                self.outer_instance.epoch_info_queue.put(epoch_info)

                if test_acc > self.outer_instance.current_val_acc and testing == False:
                    self.outer_instance.save_model()
                    self.outer_instance.current_val_acc = test_acc
                    print(f"New best model saved with validation accuracy: {test_acc}")
                    
                if test_acc > self.outer_instance.current_val_acc and testing == True:
                    self.outer_instance.current_val_acc = test_acc

        custom_callback = CustomCallback(self)
        
        final_epochs = epochs if testing else self.hyperparameters["epochs"]
        
        history = self.model.fit(
            self.X_train, self.y_train,
            epochs=final_epochs,
            batch_size=batch_size,
            validation_data=(self.X_test, self.y_test),
            callbacks=[custom_callback],
            verbose=0
        )
        
        return history

    def evaluate_model(self):
        y_pred_proba = self.model.predict(self.X_test)
        if self.is_binary:
            y_pred = (y_pred_proba > 0.5).astype(int)
        else:
            y_pred = np.argmax(y_pred_proba, axis=1)
        self.accuracy = accuracy_score(self.y_test, y_pred)
        return self.accuracy

    def save_model(self):
        if not os.path.exists(self.directory_path):
            os.makedirs(self.directory_path)
            
        self.model.save(self.complete_model_path)
        joblib.dump(self.scaler, self.complete_scaler_path)
        joblib.dump(self.label_encoders, self.complete_label_encoder_path)

    def hyperparameter_tuning(self, param_grid):
        best_accuracy = 0
        best_params = None

        param_combinations = list(product(*param_grid.values()))
        total_combinations = len(param_combinations)
        
        print(f"Starting hyperparameter tuning with {total_combinations} combinations")
        
        for i, params in enumerate(param_combinations, 1):
            params_dict = dict(zip(param_grid.keys(), params))
            print(f"Training combination {i}/{total_combinations}: {params_dict}")
            
            self.train_model(**params_dict, testing=True)
            accuracy = self.evaluate_model()

            if accuracy > best_accuracy:
                best_accuracy = accuracy
                best_params = params_dict
                print(f"New best accuracy: {best_accuracy:.4f}")

        print(f"Best accuracy: {best_accuracy:.4f} with parameters: {best_params}")
        return best_params

    def execute(self):
        param_grid = {
            "batch_size": [32],
            "epochs": [2],
            "dropout_rate": [0.15, 0.3, 0.45],
            "learning_rate": [0.01, 0.001],
            "dense_units": [(1024, 512, 128, 64, 32), (512, 128, 64, 32), 
                          (256, 128, 64, 32), (128, 64, 32)]  
        }
        # param_grid = {
        #     "batch_size": [32],
        #     "epochs": [2],
        #     "dropout_rate": [0.15],
        #     "learning_rate": [0.01, 0.001],
        #     "dense_units": [(128, 64, 32)]  
        # }

        best_params = self.hyperparameter_tuning(param_grid)
        print(f"Training final model with best parameters: {best_params}")
        
        history = self.train_model(**best_params)
        final_accuracy = self.evaluate_model()
        print(f"Final model accuracy: {final_accuracy:.4f}")
        
        data = list(self.df.drop(columns=[self.target_col]).values)[0] 
        
        formatted_dat = [f"'{item}'" if isinstance(
            item, str) else str(item) for item in data]
        
        
        interference_code = f'''
import os
import pandas as pd
import numpy as np
import joblib
from tensorflow.keras.models import load_model

# Load necessary components from local storage
def load_model_from_local(model_path):
    return load_model(model_path)

def load_scaler_from_local(scaler_path):
    return joblib.load(scaler_path)

def load_label_encoders_from_local(label_encoder_path):
    return joblib.load(label_encoder_path)

# Preprocess input data
def preprocess_input(data, scaler, label_encoders, column_names):
    df = pd.DataFrame([data], columns=column_names)
    
    for column, le in label_encoders.items():
        df[column] = le.transform(df[column])
    
    data_scaled = scaler.transform(df)
    return data_scaled

# Make predictions
def make_predictions(model, data_scaled):
    predictions_proba = model.predict(data_scaled)
    if predictions_proba.shape[1] == 1:  # Binary classification
        predictions = (predictions_proba > 0.5).astype(int)
    else:  # Multi-class classification
        predictions = np.argmax(predictions_proba, axis=1)
    return predictions, predictions_proba

def infer(input_data, model_name, scaler_name, label_encoder_name, dataset_path):
    model_path = model_name
    scaler_path = scaler_name
    label_encoder_path = label_encoder_name

    model = load_model_from_local(model_path)
    scaler = load_scaler_from_local(scaler_path)
    label_encoders = load_label_encoders_from_local(label_encoder_path)

    if model and scaler:
        df = pd.read_csv(dataset_path, encoding='{self.encoding}')
        target_idx = df.columns.get_loc('{self.target_col}')
        column_names = df.drop(columns=['{self.target_col}']).columns.tolist()

        
        y = df['{self.target_col}']

        data_scaled = preprocess_input(input_data, scaler, label_encoders, column_names)

        predictions, predictions_proba = make_predictions(model, data_scaled)

        if predictions_proba.shape[1] == 1:  # Binary classification
            pred = None
            if y.dtype == object:  
                    pred = label_encoders["target"].inverse_transform(predictions)[0]
            else:
                    pred = int(predictions[0][0])
            return {{
                    "prediction": [
                        {{"predicted_class": pred}},
                        {{"probability": float(predictions_proba[0][0])}}
                    ]
                }}    
        else:  # Multi-class classification
            if y.dtype == object: 
                pred = label_encoders["target"].inverse_transform([predictions[0]])[0]
                return {{
                        "prediction": [
                            {{"predicted_class": pred}},
                            {{"probabilities": predictions_proba[0].tolist()}}
                        ]
                    }}    
            else:
                return {{
                        "prediction": [
                           {{"predicted_class": [predictions[0]][0]}},
                            {{"probabilities": predictions_proba[0].tolist()}}
                        ]
                    }}  

    else:
        raise ValueError("Failed to load model or scaler.")

if __name__ == "__main__":
    input_data = [{', '.join(formatted_dat)}]
    model_name = '{self.complete_model_path}'
    scaler_name = '{self.complete_scaler_path}'
    label_encoder_name = '{self.complete_label_encoder_path}'
    dataset_url = '{self.dataset_url}'

    result = infer(input_data, model_name, scaler_name, label_encoder_name, dataset_url)
    print(result)
'''
        with open(f'inference{self.model_name}.py', 'w') as file:
            file.write(interference_code)

        return self.epoch_data



from TextModel import TextModel
import json

def returnArch (data, task, mainType, archType):
    current_task = data[task]

    for i in current_task:
        if  i["type"] == mainType and i["archType"] == archType:
            return i["architecture"], i["hyperparameters"]

arch_data = {}
task = "text"
mainType='topic classification'
archType='default'

with open ('arch.json', 'r') as f:
    arch_data = json.load(f)

architecture, hyperparameters = returnArch(arch_data, task, mainType, archType)

model = TextModel(
    dataset_url='bbc-text.csv',
    hasChanged=False,
    task='text',
    mainType='topic classification',
    archType='default',
    architecture=architecture,
    hyperparameters=hyperparameters
)

executor = model.execute()
for epoch_info in executor:
    if isinstance(epoch_info, dict) and 'epoch' in epoch_info:
        print(f"Epoch {epoch_info['epoch']}: Train Loss: {epoch_info['train_loss']:.4f}, "
              f"Train Acc: {epoch_info['train_acc']:.4f}, Val Loss: {epoch_info['val_loss']:.4f}, "
              f"Val Acc: {epoch_info['val_acc']:.4f}")
    else:
        print("Final model object:", epoch_info)
        break


# model_trainer.py
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from xgboost import XGBClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
import joblib
import requests
import os
import uuid


class ClassificationML:
    def __init__(self, dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters):
        self.dataset_url = dataset_url
        self.hasChanged = hasChanged
        self.task = task
        self.mainType = mainType
        self.archType = archType
        self.architecture = architecture
        self.hyperparameters = hyperparameters
        self.api_url = 'https://s3-api-uat.idesign.market/api/upload'
        self.bucket_name = 'idesign-quotation'

        self.load_data()
        self.preprocess_data()

    def load_data(self):
        self.df = pd.read_csv(self.dataset_url)
        self.X = self.df.iloc[:, :-1]
        self.y = self.df.iloc[:, -1]

    def preprocess_data(self):
        self.scaler = StandardScaler()
        self.X_standardized = self.scaler.fit_transform(self.X)
        self.X = self.X_standardized
        self.y = self.y.values
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            self.X, self.y, test_size=0.2, random_state=42)

    def create_model(self):
        if self.task == "classification":
            if self.archType == '2':
                print("Using XGB")
                self.model = XGBClassifier(**self.hyperparameters)
            elif self.archType == "1":
                print("Using SVC")
                self.model = SVC(**self.hyperparameters)
            elif self.archType == "3":
                print("Using DecisionTreeClassifier")
                self.model = DecisionTreeClassifier(**self.hyperparameters)
            else:
                raise ValueError(f"Unsupported model type: {self.mainType}")

    def train_model(self):
        self.model.fit(self.X_train, self.y_train)

    def evaluate_model(self):
        y_pred = self.model.predict(self.X_test)
        self.accuracy = accuracy_score(self.y_test, y_pred)
        print(f"Accuracy: {self.accuracy}")

    def save_model(self):
        model_filename = f'model{str(uuid.uuid4())}.pkl'
        scaler_filename = f'scaler{str(uuid.uuid4())}.pkl'

        joblib.dump(self.model, model_filename)
        joblib.dump(self.scaler, scaler_filename)

        self.model_path = model_filename
        self.scaler_path = scaler_filename

    def upload_files_to_api(self):
        try:
            files = {
                'bucketName': (None, self.bucket_name),
                'files': open(self.model_path, 'rb')
            }
            response_model = requests.put(self.api_url, files=files)
            response_data_model = response_model.json()
            model_url = response_data_model.get('locations', [])[
                0] if response_model.status_code == 200 else None

            if model_url:
                print(f"Model uploaded successfully. URL: {model_url}")
            else:
                print(
                    f"Failed to upload model. Error: {response_data_model.get('error')}")
                return None, None

            files = {
                'bucketName': (None, self.bucket_name),
                'files': open(self.scaler_path, 'rb')
            }
            response_scaler = requests.put(self.api_url, files=files)
            response_data_scaler = response_scaler.json()
            scaler_url = response_data_scaler.get(
                'locations', [])[0] if response_scaler.status_code == 200 else None

            if scaler_url:
                print(f"Scaler uploaded successfully. URL: {scaler_url}")
            else:
                print(
                    f"Failed to upload scaler. Error: {response_data_scaler.get('error')}")
                return model_url, None

            return model_url, scaler_url

        except requests.exceptions.RequestException as e:
            print(f"An error occurred: {str(e)}")
            return None, None

    def execute(self):
        self.create_model()
        self.train_model()
        self.evaluate_model()
        self.save_model()
        model_url, scaler_url = self.upload_files_to_api()

        if model_url and scaler_url:
            _id = str(uuid.uuid4())
            model_obj = {
                "modelUrl": model_url,
                "size": os.path.getsize(self.model_path) / (1024 ** 3),
                "id": _id,
                "helpers": [{"scaler": scaler_url}],
                "modelArch": self.architecture,
                "hyperparameters": self.hyperparameters,
                "task": self.task,
                "datasetUrl": self.dataset_url
            }
            return model_obj
        else:
            return None



import requests
import torch
import zipfile
from sentence_transformers import SentenceTransformer
import os
import re
import nltk
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from sentence_transformers import util
import requests
import json

model_zip_url = 'https://idesign-quotation.s3.ap-south-1.amazonaws.com/NO_COMPANYNAME/sentence_transformer_model.zip'

extract_folder = './sentence_transformer_model'

os.makedirs(extract_folder, exist_ok=True)

print("Downloading model zip file...")
response = requests.get(model_zip_url, stream=True)
zip_file_path = './sentence_transformer_model.zip'

with open(zip_file_path, 'wb') as file:
    for chunk in response.iter_content(chunk_size=1024):
        if chunk:
            file.write(chunk)

print("Unzipping model...")
with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
    zip_ref.extractall(extract_folder)

model_path = os.path.join('sentence_transformer_model')
model = SentenceTransformer(model_path)

question_embeddings_url = 'https://idesign-quotation.s3.ap-south-1.amazonaws.com/NO_COMPANYNAME/question_embeddings.pt'
answer_embeddings_url = 'https://idesign-quotation.s3.ap-south-1.amazonaws.com/NO_COMPANYNAME/answer_embeddings.pt'

response = requests.get(question_embeddings_url)
with open('question_embeddings.pt', 'wb') as file:
    file.write(response.content)

response = requests.get(answer_embeddings_url)
with open('answer_embeddings.pt', 'wb') as file:
    file.write(response.content)

question_embeddings = torch.load('question_embeddings.pt')
answer_embeddings = torch.load('answer_embeddings.pt')

nltk.download('punkt')
nltk.download('stopwords')

data_url = 'https://idesign-quotation.s3.ap-south-1.amazonaws.com/NO_COMPANYNAME/data_new.json'

response = requests.get(data_url)
qa_data = json.loads(response.text)

questions = [item['question'] for item in qa_data]
answers = [item['answer'] for item in qa_data]

questions = [item['question'] for item in qa_data]
answers = [item['answer'] for item in qa_data]

stop_words = set(stopwords.words('english'))

def preprocess_text(text):
    text = text.lower()
    text = re.sub(r'[^a-zA-Z0-9\s]', ' ', text)  
    text = re.sub(r'\s+', ' ', text).strip()  
    tokens = word_tokenize(text)
    tokens = [word for word in tokens if word not in stop_words]
    return ' '.join(tokens)

def get_answer(question):
    processed_question = preprocess_text(question)
    question_embedding = model.encode(processed_question, convert_to_tensor=True)
    
    similarities = util.pytorch_cos_sim(question_embedding, question_embeddings)[0]
    top_results = similarities.topk(k=5)
    print(top_results)
    similarity, index = similarities.max(), similarities.argmax()
    similarity_percentage = similarity.item() * 100
    
    if similarity_percentage > 45:
        return answers[index], similarity_percentage
    else:
        return "Sorry, I didn't understand that!", similarity_percentage

user_question = "What is munafa"
answer, similarity_percentage = get_answer(user_question)

print(f"Answer: {answer}")
print(f"Similarity Percentage: {similarity_percentage:.2f}%")


import fitz
import google.generativeai as genai
import os
import uuid

api_key = "AIzaSyC93XxpL8z7dz4UjNBvECFYaobAOQre0Bk"
genai.configure(api_key=api_key)

class ChatbotPDF:
    def __init__(self, dataset_url, userId, task):
        self.dataset_url = dataset_url
        self.pdf_dir = 'pdfs'
        self.pdf_name = self.dataset_url.split('/')[-1]
        self.pdf_path = os.path.join(self.pdf_dir, self.pdf_name)
        self.userId = userId
        self._id = str(uuid.uuid4())
        self.task = task

    # def extract(self):
    # self.pdf_name = self.dataset_url.split('/')[-1]
    #     self.pdf_path = os.path.join(self.pdf_dir, self.pdf_name)
    #     pdf_document = fitz.open(self.pdf_path)

    #     text = ""
    #     for page_num in range(pdf_document.page_count):
    #         page = pdf_document.load_page(page_num)
    #         text += page.get_text()

    #     pdf_document.close()

    #     return text

    def execute(self):
        # if not os.path.exists(self.pdf_path):
        #     raise FileNotFoundError(f"The file {self.pdf_name} does not exist in the directory {self.pdf_dir}")
        
        # text = self.extract()

        # model = genai.GenerativeModel("gemini-1.5-flash")
        # response = model.generate_content(f"Context: {text} Answer the following question in less than 100 words no matter what and if the answer doesnt exist in the context, simple reply with answer not available: {self.question}")

        # print(response.text)
        interference_code = ""

#         api_code_python = f'''
# import requests
# import json

# url = "https://apiv3.xanderco.in/core/interference/" 

# data = {{
#     "data": "Your input text",
#     "modelId": '{self._id}',
#     "userId": '{self.userId}',
# }}

# try:
#     response = requests.post(url, json=data)

#     if response.status_code == 200:
#         # Print the response JSON
#         print("Response:")
#         print(json.dumps(response.json(), indent=2))
#     else:
#         print(f"Error: {{response.status_code}}")
#         print(response.text)
# except requests.exceptions.RequestException as e:
#     print(f"An error occurred: {{e}}")
# '''

        api_code_python = f'''
<script src="https://chatbot.xanderco.in/chatbot-loader.js?modelId={self._id}&userId={self.userId}"></script>
        '''

        api_code_js = f'''
const url = "https://apiv3.xanderco.in/core/interference/";

const data = {{
    data: "Your text here",
    modelId: '{self._id}',
    userId: '{self.userId}',
}};

const headers = {{
    'Content-Type': 'application/json',
}};

fetch(url, {{
    method: 'POST',
    headers: headers,
    body: JSON.stringify(data)
}})
.then(response => response.json().then(data => {{
    if (response.ok) {{
        console.log("Response:");
        console.log(JSON.stringify(data, null, 2));
    }} else {{
        console.error(`Error: ${{response.status}}`);
        console.error(data);
    }}
}}))
.catch(error => {{
    console.error(`An error occurred: ${{error}}`);
}});
'''
        model_obj = {
            "modelUrl": "",
            "helpers": [],
            "id": self._id,
            "architecture": "Language Models",
            "hyperparameters": {},
            "size": 0,
            "interferenceCode": interference_code,
            "datasetUrl": self.dataset_url,
            "codes": [
                    {"python": api_code_python},
                    {"javascript": api_code_js}
            ],
            "task": self.task,
        }

        return model_obj


# aiunified/__init__.py

from .RegressionDL import RegressionDL
from .ClassificationDL import ClassificationDL
from .ImageModelTrainer import ImageModelTrainer
from .TextModel import TextModel

__all__ = [
    "RegressionDL",
    "ClassificationDL",
    "Forecaster",
    "ImageModelTrainer",
    "TextModel",
]



import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import train_test_split
import tensorflow as tf
from tensorflow.keras.preprocessing.text import Tokenizer
from tensorflow.keras.preprocessing.sequence import pad_sequences
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Embedding, LSTM, Dense, Dropout, Bidirectional
from tensorflow.keras.regularizers import l2
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, Callback
from tensorflow.keras.utils import to_categorical
import nltk
import os
import pickle
import queue
import itertools
import chardet

class TextModel:
    def __init__(self, dataset_url, model_name, default_hyperparameters=None, target_col="", encoding=""):
        physical_devices = tf.config.list_physical_devices('GPU')
        if physical_devices:
            print("GPU(s) detected:", len(physical_devices))
            for device in physical_devices:
                try:
                    tf.config.experimental.set_memory_growth(device, True)
                    print(f"Memory growth enabled for GPU device: {device}")
                    
                except RuntimeError as e:
                    print(f"Error configuring GPU device {device}: {e}")
            
            tf.keras.mixed_precision.set_global_policy('mixed_float16')
            print("Mixed precision policy enabled for GPU training")
        else:
            print("No GPU devices detected. Running on CPU.")
        self.dataset_url = dataset_url
        self.model_name = model_name
        self.epochs = default_hyperparameters["epochs"]
        self.default_hyperparameters = default_hyperparameters or {
            "batch_size": 32,
            "epochs": 10,
            "embedding_dim": 100,
            "lstm_units": 128,
            "dropout_rate": 0.5,
            "learning_rate": 0.001,
        }
        self.best_hyperparameters = self.default_hyperparameters
        self.epoch_info_queue = queue.Queue()
        self.directory_path = "models"
        self.current_val_acc = 0
        self.epoch_data = []
        self.model_path = os.path.join(self.directory_path, f'{self.model_name}.h5')
        self.tokenizer_path = os.path.join(self.directory_path, f'tokenizer_{self.model_name}.pkl')
        self.label_encoder_path = os.path.join(self.directory_path, f'label_encoder_{self.model_name}.pkl')
        self.target_col = target_col
        self.encoding = encoding
        self.load_data()
        self.preprocess_data()

    def load_data(self):
        with open(self.dataset_url, 'rb') as file:
            result = chardet.detect(file.read())
            encoding = result['encoding']
        
        self.df = pd.read_csv(self.dataset_url, encoding=encoding)        
        self.df = self.df.dropna()

    def preprocess_data(self):
        target_idx = self.df.columns.get_loc(self.target_col)
        
        text_cols = self.df.drop(columns=[self.target_col])
        self.text_columns = text_cols.apply(lambda x: ' '.join(x.astype(str)), axis=1)
        
        self.category_column = self.target_col
        
        self.label_encoder = LabelEncoder()
        self.df[self.category_column] = self.label_encoder.fit_transform(self.df[self.category_column])
        self.num_classes = len(np.unique(self.df[self.category_column]))
        
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            self.text_columns, self.df[self.category_column], test_size=0.2, random_state=42
        )
        
        self.max_num_words = 20000
        self.max_sequence_length = 100
        self.tokenizer = Tokenizer(num_words=self.max_num_words)
        self.tokenizer.fit_on_texts(self.X_train)
        self.X_train_seq = self.tokenizer.texts_to_sequences(self.X_train)
        self.X_test_seq = self.tokenizer.texts_to_sequences(self.X_test)
        self.X_train_pad = pad_sequences(self.X_train_seq, maxlen=self.max_sequence_length)
        self.X_test_pad = pad_sequences(self.X_test_seq, maxlen=self.max_sequence_length)
        
        self.y_train_cat = to_categorical(self.y_train, self.num_classes)
        self.y_test_cat = to_categorical(self.y_test, self.num_classes)
        
    def create_model(self, hyperparameters):
        model = Sequential()
        model.add(Embedding(
            input_dim=self.max_num_words,
            output_dim=hyperparameters["embedding_dim"],
            input_length=self.max_sequence_length
        ))
        model.add(Bidirectional(LSTM(
            units=hyperparameters["lstm_units"],
            dropout=hyperparameters["dropout_rate"],
            recurrent_dropout=hyperparameters["dropout_rate"],
            return_sequences=False
        )))
        model.add(Dropout(hyperparameters["dropout_rate"]))
        model.add(Dense(
            units=self.num_classes,
            activation='softmax',
            kernel_regularizer=l2(0.01)
        ))

        optimizer = tf.keras.optimizers.Adam(learning_rate=hyperparameters["learning_rate"])
        model.compile(optimizer=optimizer, loss='categorical_crossentropy', metrics=['accuracy'])
        return model

    def train_model(self, hyperparameters, testing=False):
        model = self.create_model(hyperparameters)
        batch_size = hyperparameters["batch_size"]
        epochs = hyperparameters["epochs"]
        
        if testing == False:
            self.current_val_acc = 0
            
        print(self.current_val_acc)
        class CustomCallback(Callback):
            def __init__(self, outer_instance):
                super().__init__()
                self.outer_instance = outer_instance

            def on_epoch_end(self, epoch, logs=None):
                test_acc = logs.get('val_accuracy')
                epoch_info = {
                    "epoch": epoch + 1,
                    "train_loss": logs.get('loss'),
                    "train_acc": logs.get('accuracy'),
                    "test_loss": logs.get('val_loss'),
                    "test_acc": logs.get('val_accuracy'),
                }
                print(epoch_info)
                self.outer_instance.epoch_data.append(epoch_info)
                self.outer_instance.epoch_info_queue.put(epoch_info)

                if test_acc > self.outer_instance.current_val_acc and testing == False:
                    self.outer_instance.save_model(model)
                    self.outer_instance.current_val_acc = test_acc
                    print(f"New best model saved with validation accuracy: {test_acc:.4f}")
                elif test_acc > self.outer_instance.current_val_acc and testing == True:
                    self.outer_instance.current_val_acc = test_acc

        custom_callback = CustomCallback(self)

        final_epochs = 0
        
        if testing:
            final_epochs = epochs
        else:
            final_epochs = self.epochs
            
        history = model.fit(
            self.X_train_pad, self.y_train_cat,
            batch_size=batch_size,
            epochs=final_epochs,
            validation_split=0.2,
            callbacks=[custom_callback],
            verbose=0
        )
        return history.history["val_accuracy"][-1]

    def tune_hyperparameters(self, param_grid):
        print(param_grid)
        param_names = list(param_grid.keys())
        param_combinations = list(itertools.product(*param_grid.values()))

        best_val_acc = 0
        best_params = None

        for i in range(0, len(param_combinations)):
            hyperparameters = dict(zip(param_names, param_combinations[i]))
            print(f"Testing hyperparameters {i}/{len(param_combinations)}: {hyperparameters}")
            val_acc = self.train_model(hyperparameters, testing=True)
            if val_acc >= best_val_acc:
                best_val_acc = val_acc
                best_params = hyperparameters
                self.best_hyperparameters = hyperparameters
                print(f"New best hyperparameters found: {best_params} with accuracy {best_val_acc:.4f}")

        print(f"Best hyperparameters: {best_params}")
        return best_params

    def save_model(self, model):
        if not os.path.exists(self.directory_path):
            os.makedirs(self.directory_path)

        model_path = os.path.join(self.directory_path, f'{self.model_name}.h5')
        tokenizer_path = os.path.join(self.directory_path, f'tokenizer_{self.model_name}.pkl')
        label_encoder_path = os.path.join(self.directory_path, f'label_encoder_{self.model_name}.pkl')

        model.save(model_path)

        with open(tokenizer_path, 'wb') as f:
            pickle.dump(self.tokenizer, f)

        with open(label_encoder_path, 'wb') as f:
            pickle.dump(self.label_encoder, f)

    def execute(self):
        nltk.download('punkt')
        param_grid = {
            "batch_size": [32],
            "epochs": [3],
            "embedding_dim": [50, 100, 150],
            "lstm_units": [64, 128, 192],
            "dropout_rate": [0.1, 0.3, 0.5],
            "learning_rate": [0.01, 0.001],
        }
        # param_grid = {
        #     "batch_size": [32],
        #     "epochs": [2],
        #     "embedding_dim": [50],
        #     "lstm_units": [64],
        #     "dropout_rate": [0.1],
        #     "learning_rate": [0.01, 0.001, 0.0001],
        # }
        params = self.tune_hyperparameters(param_grid)
        print('----training begins----')
        self.train_model(params, testing=False)
            
        interference_code = f'''
import os
import numpy as np
import tensorflow as tf
from tensorflow.keras.models import load_model
from tensorflow.keras.preprocessing.sequence import pad_sequences
import joblib

# Load necessary components from local storage
def load_model_from_local(model_path):
    return load_model(model_path)

def load_tokenizer(tokenizer_path):
    return joblib.load(tokenizer_path)

def load_label_encoder(label_encoder_path):
    return joblib.load(label_encoder_path)

# Preprocess input text
def preprocess_text(text, tokenizer, max_sequence_length):
    sequences = tokenizer.texts_to_sequences([text])
    padded_sequences = pad_sequences(sequences, maxlen=max_sequence_length)
    return padded_sequences

# Make predictions
def make_predictions(model, preprocessed_text):
    predictions = model.predict(preprocessed_text)
    return predictions

def infer(input_data, model_name, tokenizer_name, label_encoder_name):
    model_path = model_name
    tokenizer_path = tokenizer_name
    label_encoder_path = label_encoder_name
    
    # Load model and preprocessing components
    model = load_model_from_local(model_path)
    tokenizer = load_tokenizer(tokenizer_path)
    label_encoder = load_label_encoder(label_encoder_path)
    
    if model and tokenizer and label_encoder:
        # Set maximum sequence length for padding
        max_sequence_length = 100  # Adjust this based on your model's requirements
        
        # Preprocess the input text
        preprocessed_text = preprocess_text(input_data, tokenizer, max_sequence_length)
        
        # Get predictions
        predictions_proba = make_predictions(model, preprocessed_text)
        predicted_class = tf.argmax(predictions_proba, axis=1)
        predicted_label = label_encoder.inverse_transform(predicted_class)
        
        # Prepare output
        return {{
            "prediction": [
                {{"predicted_class": predicted_label[0]}},
                {{"probabilities": predictions_proba[0].tolist()}}
            ]
        }}
    else:
        raise ValueError("Failed to load model, tokenizer, or label encoder.")

if __name__ == "__main__":
    input_data = 'test'  # Your input text here
    model_name = '{self.model_path}'
    tokenizer_name = '{self.tokenizer_path}'
    label_encoder_name = '{self.label_encoder_path}'
    
    result = infer(input_data, model_name, tokenizer_name, label_encoder_name)
    print(result)
        '''
        with open(f'inference{self.model_name}.py', 'w') as file:
            file.write(interference_code)
        yield 'Done'




import tensorflow as tf
from tensorflow import keras
from keras import layers, models, callbacks, regularizers
from tensorflow.keras.utils import image_dataset_from_directory
import os
import zipfile
import queue
import keras_tuner as kt

class ImageModelTrainer:
    def __init__(self, dataset_url, model_name, hyperparameters=None):
        physical_devices = tf.config.list_physical_devices('GPU')
        if physical_devices:
            print("GPU(s) detected:", len(physical_devices))
            for device in physical_devices:
                try:
                    tf.config.experimental.set_memory_growth(device, True)
                    print(f"Memory growth enabled for GPU device: {device}")
                    
                except RuntimeError as e:
                    print(f"Error configuring GPU device {device}: {e}")
            
            tf.keras.mixed_precision.set_global_policy('mixed_float16')
            print("Mixed precision policy enabled for GPU training")
        else:
            print("No GPU devices detected. Running on CPU.")
        self.dataset_url = dataset_url
        self.model_name = model_name
        self.task = 'image'
        self.hyperparameters = hyperparameters or {}
        self.epoch_info_queue = queue.Queue()
        self.data_dir = "extracted_files"
        self.img_height = 150
        self.img_width = 150
        self.epoch_data = []
        self.model_path = f"{model_name}.keras"
        self.directory_path = "models"
        self.complete_path = os.path.join(self.directory_path, self.model_path)
        self.train_ds = None
        self.val_ds = None
        self.num_classes = 0
        self.download_and_extract_data()
        self.prepare_datasets()

    def download_and_extract_data(self):
        with zipfile.ZipFile(self.dataset_url, "r") as zip_ref:
            zip_ref.extractall(self.data_dir)
        self.data_dir = os.path.join(self.data_dir, os.listdir(self.data_dir)[0])
        print(f"Files extracted to: {self.data_dir}")

    def prepare_datasets(self):
        train_ds = image_dataset_from_directory(
            self.data_dir,
            validation_split=0.2,
            subset="training",
            seed=123,
            image_size=(self.img_height, self.img_width),
            batch_size=self.hyperparameters.get("batch_size", 32),
        )
        val_ds = image_dataset_from_directory(
            self.data_dir,
            validation_split=0.2,
            subset="validation",
            seed=123,
            image_size=(self.img_height, self.img_width),
            batch_size=self.hyperparameters.get("batch_size", 32),
        )

        AUTOTUNE = tf.data.AUTOTUNE
        self.class_names = train_ds.class_names
        self.num_classes = len(self.class_names)
        print(f"Class names: {self.class_names}")

        self.train_ds = train_ds.cache().shuffle(1000).prefetch(buffer_size=AUTOTUNE)
        self.val_ds = val_ds.cache().prefetch(buffer_size=AUTOTUNE)

    def model_builder(self, hp):
        model = keras.Sequential()
        conv_activation = hp.Choice('conv_activation', values=['relu', 'tanh', 'leaky_relu'])
        dense_activation = hp.Choice('dense_activation', values=['relu', 'tanh', 'leaky_relu'])
        regularizer_choice = hp.Choice('regularizer', values=['l1', 'l2', 'l1_l2'])

        def get_regularizer():
            if regularizer_choice == 'l1':
                return regularizers.l1(hp.Float('l1', 1e-6, 1e-2, sampling='log'))
            elif regularizer_choice == 'l2':
                return regularizers.l2(hp.Float('l2', 1e-6, 1e-2, sampling='log'))
            return regularizers.l1_l2(
                l1=hp.Float('l1', 1e-6, 1e-2, sampling='log'),
                l2=hp.Float('l2', 1e-6, 1e-2, sampling='log'),
            )

        model.add(layers.Rescaling(1.0 / 255, input_shape=(self.img_height, self.img_width, 3)))

        for i in range(hp.Int('num_conv_layers', 1, 4)):
            model.add(layers.Conv2D(
                filters=hp.Int(f'conv_{i}_filters', 32, 128, step=32),
                kernel_size=(3, 3),
                activation=conv_activation,
                kernel_regularizer=get_regularizer()
            ))
            model.add(layers.MaxPooling2D(pool_size=(2, 2)))

        model.add(layers.Flatten())

        for i in range(hp.Int('num_dense_layers', 1, 3)):
            model.add(layers.Dense(
                units=hp.Int(f'dense_{i}_units', 64, 256, step=64),
                activation=dense_activation,
                kernel_regularizer=get_regularizer()
            ))
            model.add(layers.Dropout(hp.Float(f'dropout_{i}', 0.1, 0.5, step=0.15)))

        model.add(layers.Dense(self.num_classes))

        model.compile(
            optimizer=keras.optimizers.Adam(hp.Float('learning_rate', 1e-4, 1e-2, sampling='log')),
            loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
            metrics=['accuracy']
        )

        return model

    def tune_hyperparameters(self):
        tuner = kt.Hyperband(
            self.model_builder,
            objective='val_accuracy',
            max_epochs=20,
            factor=3,
            directory='tuned_models',
            project_name=self.model_name
        )

        early_stopping = callbacks.EarlyStopping(monitor="val_accuracy", patience=5, restore_best_weights=True)

        tuner.search(
            self.train_ds,
            validation_data=self.val_ds,
            epochs=20,
            callbacks=[early_stopping]
        )

        best_hps = tuner.get_best_hyperparameters(num_trials=1)[0]
        print(f"Best hyperparameters: {best_hps.values}")

        return best_hps

    def build_and_train_final_model(self, best_hps):
        self.model = self.model_builder(best_hps)
        early_stopping = callbacks.EarlyStopping(monitor="val_accuracy", patience=5, restore_best_weights=True)

        history = self.model.fit(
            self.train_ds,
            validation_data=self.val_ds,
            epochs=self.hyperparameters["epochs"],
            callbacks=[early_stopping],
            verbose=1
        )

        if not os.path.exists(self.directory_path):
            os.makedirs(self.directory_path)
        self.model.save(self.complete_path)
        print(f"Model saved to {self.complete_path}")

        return history

    def execute(self):
        best_hps = self.tune_hyperparameters()
        history = self.build_and_train_final_model(best_hps)
        
        interference_code = f'''
import os
import numpy as np
import tensorflow as tf
from tensorflow.keras.models import load_model
from tensorflow.keras.preprocessing import image
from PIL import Image
import requests
from io import BytesIO

# Load model from local storage
def load_model_from_local(model_path):
    return load_model(model_path)

# Prepare image for prediction
def prepare_image(img, img_height=120, img_width=120):
    img = img.resize((img_width, img_height))
    img_array = image.img_to_array(img)
    img_array = tf.expand_dims(img_array, 0)
    return img_array

# Make predictions
def make_predictions(model, img_array):
    predictions = model.predict(img_array)
    score = tf.nn.softmax(predictions[0])
    return tf.argmax(score), tf.reduce_max(score)

def load_image(image_path=None, image_url=None):
    if image_path:
        return Image.open(image_path)
    elif image_url:
        response = requests.get(image_url)
        return Image.open(BytesIO(response.content))
    else:
        raise ValueError("Please provide either an image path or an image URL")

def infer(input_data, model_name, class_names):
    try:
        # Setup model path and load model
        model_path = model_name
        model = load_model_from_local(model_path)
        
        if model:
            # Handle image input
            if isinstance(input_data, dict):
                image_path = input_data.get('image_path')
                image_url = input_data.get('image_url')
                
                if not image_path and not image_url:
                    raise ValueError("Please provide either an image path or an image URL")
                
                # Load and prepare image
                img = load_image(image_path, image_url)
                img_array = prepare_image(img)
                
                # Get predictions
                class_idx, class_prob = make_predictions(model, img_array)
                
                # Prepare output
                return {{
                    "prediction": [
                        {{"predicted_class": class_names[int(class_idx)]}},
                        {{"probability": float(class_prob)}}
                    ]
                }}
            else:
                raise ValueError("Input data should be a dictionary with 'image_path' or 'image_url'")
        else:
            raise ValueError("Failed to load model.")
            
    except Exception as e:
        raise Exception(f"Error during inference: {{str(e)}}")

if __name__ == "__main__":
    input_data = {{
        'image_path': '',  
        'image_url': ''  
    }}
    model_name = '{self.complete_path}'
    class_names = {self.class_names}
    
    result = infer(input_data, model_name, class_names)
    print(result)
        '''
        with open(f'inference{self.model_name}.py', 'w') as file:
            file.write(interference_code)
        return history



import pandas as pd
from RegressionML import RegressionML
from RegressionDL import RegressionDL
import json

def returnArch(data, task, mainType, archType):
    current_task = data[task]

    for i in current_task:
        if i["type"] == mainType and i["archType"] == archType:
            return i["architecture"], i["hyperparameters"]

if __name__ == "__main__":
    dataset_url = "test.csv"  # will be sent by user
    hasChanged = False  # will be sent by user
    task = "regression"  # will be sent by user
    mainType = "DL"  # will be sent by user
    archType = "default"  # will be sent by user
    arch_data = {}  # will be sent by user if hasChanged = true
    
    with open('arch.json', 'r') as f:
        arch_data = json.load(f)

    # print(arch_data)

    if task == "regression" and hasChanged == False:
        if mainType == "DL":
            architecture, hyperparameters = returnArch(
                arch_data, task, mainType, archType)
            model_trainer = RegressionDL(
                dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters, "9")
            print("executing")
            executor = model_trainer.execute()
            
            for epoch_info in executor:
                if isinstance(epoch_info, dict) and 'epoch' in epoch_info:
                    print(f"Epoch {epoch_info['epoch']}: Train Loss: {epoch_info['train_loss']:.4f}, Test Loss: {epoch_info['test_loss']:.4f}")
                else:
                    print("Final model object:", epoch_info)
                    break

        elif mainType == "ML":
            print("In ML")
            architecture, hyperparameters = returnArch(
                arch_data, task, mainType, archType)
            model_trainer = RegressionML(
                dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters, "9")
            model_obj = model_trainer.execute()
            print(model_obj)
    if task == "regression" and hasChanged == True:
        if mainType == "DL":
            architecture = [
                {
                    "layer": "Dense",
                    "neurons": 128,
                    "activation": "relu",
                    "define_input_shape": "true"
                },
                {"layer": "Dropout", "ratio": 0.1},
                {
                    "layer": "Dense",
                    "neurons": 64,
                    "activation": "relu",
                    "define_input_shape": "false"
                },
                {"layer": "Dropout", "ratio": 0.1},
                {
                    "layer": "Dense",
                    "neurons": 32,
                    "activation": "relu",
                    "define_input_shape": "false"
                },
                {"layer": "Dense", "neurons": 1, "define_input_shape": "false"}
            ]

            hyperparameters = {"epochs": 1,
                               "batch_size": 32, "validation_size": 0.2}
            model_trainer = RegressionDL(
                dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters, "9")
            model_obj = model_trainer.execute()
            print(model_obj)
        elif mainType == "ML":
            print("In ML")
            architecture, hyperparameters = returnArch(
                arch_data, task, mainType, archType)
            model_trainer = RegressionML(
                dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters, "9")
            model_obj = model_trainer.execute()
            print(model_obj)



import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Conv2D, Flatten, LSTM, Attention, Input
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.callbacks import Callback
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
import joblib
import requests
import os
import uuid
import queue
import threading
import random
from sklearn.model_selection import GridSearchCV
from sklearn.preprocessing import LabelEncoder
from itertools import product

class RegressionDL:
    def __init__(self, dataset_url, architecture, hyperparameters, model_name, target_col, encoding):
        physical_devices = tf.config.list_physical_devices('GPU')
        if physical_devices:
            print("GPU(s) detected:", len(physical_devices))
            for device in physical_devices:
                try:
                    tf.config.experimental.set_memory_growth(device, True)
                    print(f"Memory growth enabled for GPU device: {device}")
                except RuntimeError as e:
                    print(f"Error configuring GPU device {device}: {e}")
        else:
            print("No GPU devices detected. Running on CPU.")
            
        self.dataset_url = dataset_url
        self.architecture = architecture
        self.hyperparameters = hyperparameters
        self.task_type = 'regression'
        self.model = None
        self.scaler = StandardScaler()
        self.model_path = f'{model_name}.h5'
        self.scaler_path = f'scaler_{model_name}.pkl'
        self.label_encoder_path = f'label_encoder_{model_name}.pkl'
        self.directory_path = "models"
        self.complete_model_path = os.path.join(self.directory_path, self.model_path)
        self.complete_scaler_path = os.path.join(self.directory_path, self.scaler_path)
        self.complete_label_encoder_path = os.path.join(self.directory_path, self.label_encoder_path)
        self.epoch_info_queue = queue.Queue()
        self.model_name = model_name
        self.target_col = target_col
        self.encoding = encoding
        self.current_epoch_info = None
        self.epoch_data = []
        self.current_val_loss = float('inf')
        self.data, self.label_encoder = self.load_and_prepare_data()

    def apply_preprocessing(self, value, data):
        uni = list(set(data))
        return uni.index(value)

    def load_and_prepare_data(self):
        df = pd.read_csv(self.dataset_url, encoding=self.encoding)
        df = df.dropna()
        df = df.iloc[:25000]
        
        label_encoders = {}
        
        for column_name in df.columns:
            if df[column_name].dtype in ['object']:
                le = LabelEncoder()
                df[column_name] = le.fit_transform(df[column_name])
                label_encoders[column_name] = le
                print(f"Label encoding applied to column '{column_name}'.")
        
        return df, label_encoders

    def build_model(self, params=None):
        if tf.config.list_physical_devices('GPU'):
            tf.keras.mixed_precision.set_global_policy('mixed_float16')
            print("Mixed precision policy enabled for GPU training")
            
        input_shape = (self.data.shape[1] - 1,)
        if params and 'layers' in params:
            self.model = self.build_dense_model(
                input_shape=input_shape,
                layers=params['layers'],
                activation=params.get('activation', 'relu'),
                learning_rate=params.get('learning_rate', 0.0005)
            )
        else:
            self.model = self.build_dense_model(input_shape=input_shape)

    def build_dense_model(self, input_shape, layers=[256, 128, 64, 32], 
                         activation='relu', output_activation=None, learning_rate=0.0005):
        model = Sequential()
        model.add(Dense(layers[0], input_shape=input_shape, activation=activation, 
                       kernel_regularizer=tf.keras.regularizers.l2(0.01)))
        model.add(tf.keras.layers.Dropout(0.4))
        
        for layer in layers[1:]:
            model.add(Dense(layer, activation=activation, 
                          kernel_regularizer=tf.keras.regularizers.l2(0.01)))
            model.add(tf.keras.layers.Dropout(0.4))
        
        model.add(Dense(1, activation=output_activation))
        
        optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
        model.compile(optimizer=optimizer, loss='mean_absolute_error', metrics=['mae'])
        
        return model

    def train_model(self, params=None, testing=False):
        # if params:
        #     self.hyperparameters.update(params)
        
        y = self.data[self.target_col]
        X = self.data.drop(columns=[self.target_col])
        # target_idx = self.data.columns.get_loc(self.target_col)
        # X = self.data.iloc[:, :target_idx]  
        # y = self.data.iloc[:, target_idx]
        
        X = self.scaler.fit_transform(X)
        
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.2, random_state=42)

        class CustomCallback(Callback):
            def __init__(self, outer_instance):
                super().__init__()
                self.outer_instance = outer_instance

            def on_epoch_end(self, epoch, logs=None):
                test_loss = logs.get('val_loss')
                train_loss = logs.get('loss')
                epoch_info = {
                    "epoch": epoch + 1,
                    "train_loss": train_loss,
                    "test_loss": test_loss
                }
                
                print(epoch_info)

                self.outer_instance.epoch_data.append(epoch_info)
                self.outer_instance.current_epoch_info = epoch_info
                self.outer_instance.epoch_info_queue.put(epoch_info)

                if test_loss < self.outer_instance.current_val_loss and testing == False:
                    self.outer_instance.save_model()
                    self.outer_instance.current_val_loss = test_loss
                    print(f"New best model saved with validation loss: {test_loss:.4f}")
                elif test_loss < self.outer_instance.current_val_loss and testing == True:
                    self.outer_instance.current_val_loss = test_loss

        custom_callback = CustomCallback(self)
        
        final_epochs = 0
        
        if testing == False:
            final_epochs = self.hyperparameters["epochs"]
        else:
            final_epochs = params["epochs"]
        
        history = self.model.fit(
            X_train, 
            y_train,
            validation_data=(X_test, y_test),
            epochs=final_epochs,
            batch_size=32,
            callbacks=[custom_callback],
            verbose=0
        )
        
        self.X_test, self.y_test = X_test, y_test
        return history

    def hyperparameter_tuning(self, param_grid):
        best_loss = float('inf')
        best_params = None
        
        param_combinations = list(product(*param_grid.values()))
        total_combinations = len(param_combinations)
        
        print(f"Starting hyperparameter tuning with {total_combinations} combinations")
        
        for i, params in enumerate(param_combinations, 1):
            params_dict = dict(zip(param_grid.keys(), params))
            print(f"\nTraining combination {i}/{total_combinations}:")
            print(params_dict)
            
            # Reset model and build with current parameters
            self.model = None
            self.build_model(params_dict)
            
            # Train model and get results
            self.train_model(params_dict, testing=True)
            current_loss = self.current_val_loss
            
            if current_loss < best_loss:
                best_loss = current_loss
                best_params = params_dict
                print(f"New best MAE loss: {best_loss:.4f}")
        
        print(f"\nTuning completed!")
        print(f"Best MAE loss: {best_loss:.4f}")
        print(f"Best parameters: {best_params}")
        return best_params

    def evaluate_model(self):
        results = self.model.evaluate(self.X_test, self.y_test, verbose=0)
        return results

    def save_model(self):
        if not os.path.exists(self.directory_path):
            os.makedirs(self.directory_path)
        
        self.model.save(self.complete_model_path)
        joblib.dump(self.scaler, self.complete_scaler_path)
        joblib.dump(self.label_encoder, self.complete_label_encoder_path)

    def execute_with_tuning(self):
        param_grid = {
            'batch_size': [32, 64],
            'epochs': [3],
            'layers': [
                [256, 128, 64, 32],
                [512, 256, 128, 64],
                [1024, 512, 256, 128, 64],
                [128, 64, 32]
            ],
            'activation': ['relu', 'leaky_relu'],
            'learning_rate': [0.01, 0.001, 0.0005]
        }
        # param_grid = {
        #     'epochs': [3],
        #     'layers': [
        #         [256, 128, 64, 32],
        #         [512, 256, 128, 64, 32],
        #     ],
        #     'activation': ['relu'],
        #     'learning_rate': [0.001]
        # }
        
        print("Starting hyperparameter tuning process...")
        best_params = self.hyperparameter_tuning(param_grid)
        
        print("\nTraining final model with best parameters...")
        self.model = None
        self.build_model(best_params)
        history = self.train_model(best_params)
        
        final_loss = self.evaluate_model()
        print(f"Final model MAE loss: {final_loss[0]:.4f}")
        
        return history

    def execute(self):
        self.build_model()
        training_thread = threading.Thread(target=self.execute_with_tuning)
        training_thread.start()

        epochs_completed = 0
        while epochs_completed < int(self.hyperparameters['epochs']):
            try:
                epoch_info = self.epoch_info_queue.get(timeout=1)
                yield epoch_info
                epochs_completed += 1
            except queue.Empty:
                if not training_thread.is_alive():
                    break

        training_thread.join()
        
        data = list(self.data.drop(columns=[self.target_col]).values)[0] 
        
        formatted_dat = [f"'{item}'" if isinstance(
            item, str) else str(item) for item in data]
        
#         interference_code = f'''import os
# import pandas as pd
# import numpy as np
# import joblib
# from tensorflow.keras.models import load_model

# # Load necessary components from local storage
# def load_model_from_local(model_path):
#     return load_model(model_path)

# def load_scaler_from_local(scaler_path):
#     return joblib.load(scaler_path)

# def load_label_encoders_from_local(label_encoder_path):
#     return joblib.load(label_encoder_path)

# # Preprocess input data
# def preprocess_input(data, scaler, label_encoders, column_names):
#     df = pd.DataFrame([data], columns=column_names)
#     for column, le in label_encoders.items():
#         if column in df.columns:
#             df[column] = le.transform(df[column])
#     df = pd.get_dummies(df, columns=df.select_dtypes(include=['object']).columns, drop_first=True)
#     df = df.reindex(columns=scaler.feature_names_in_, fill_value=0)
#     data_scaled = scaler.transform(df)
#     return data_scaled

# # Make predictions
# def make_predictions(model, data_scaled):
#     predictions = model.predict(data_scaled)
#     return predictions

# def infer(input_data, model_name, scaler_name, label_encoder_name, dataset_path):
#     model_path = model_name
#     scaler_path = scaler_name
#     label_encoder_path = label_encoder_name
    
#     model = load_model_from_local(model_path)
#     scaler = load_scaler_from_local(scaler_path)
#     label_encoders = load_label_encoders_from_local(label_encoder_path)
    
#     if model and scaler:
#         df = pd.read_csv(dataset_path, encoding='{self.encoding}')
#         column_names = df.columns.drop(['{self.target_col}']).tolist()
#         data_scaled = preprocess_input(input_data, scaler, label_encoders, column_names)
#         predictions = make_predictions(model, data_scaled)
        
#         # For regression, return the predicted values directly
#         return {{
#             "prediction": [
#                 {{"predicted_value": float(predictions[0][0])}}
#             ]
#         }}
#     else:
#         raise ValueError("Failed to load model or scaler.")

# if __name__ == "__main__":
#     input_data = [{', '.join(formatted_dat)}]
#     model_name = '{self.complete_model_path}'
#     scaler_name = '{self.complete_scaler_path}'
#     label_encoder_name = '{self.complete_label_encoder_path}'
#     dataset_url = '{self.dataset_url}'
    
#     result = infer(input_data, model_name, scaler_name, label_encoder_name, dataset_url)
#     print(result)
#         '''
        interference_code  = f'''
import numpy as np
import pandas as pd
import tensorflow as tf
import joblib
import os

def make_prediction(input_data, model_path, scaler_path, label_encoder_path, dataset_url):
    model = tf.keras.models.load_model(model_path)
    scaler = joblib.load(scaler_path)
    label_encoders = joblib.load(label_encoder_path)
    
    df_original = pd.read_csv(dataset_url, encoding='{self.encoding}')
    
    feature_columns = df_original.columns.drop(['{self.target_col}']).tolist()
    
    if len(input_data) != len(feature_columns):
        raise ValueError(f"Input data must have {{len(feature_columns)}} features. Got {{len(input_data)}} instead.")
    
    input_df = pd.DataFrame([input_data], columns=feature_columns)
    
    for column, encoder in label_encoders.items():
        if column in input_df.columns: 
            try:
                input_df[column] = encoder.transform(input_df[column])
            except ValueError:
                most_frequent_category = encoder.classes_[0]
                input_df[column] = input_df[column].map(lambda x: x if x in encoder.classes_ else most_frequent_category)
                input_df[column] = encoder.transform(input_df[column])
    
    scaled_data = scaler.transform(input_df)
    
    prediction = model.predict(scaled_data)
    
    return {{
            "prediction": [
                {{"predicted_value": float(prediction[0][0])}}
            ]
        }}

if __name__ == "__main__":
    input_data = [{', '.join(formatted_dat)}]
    model_name = '{self.complete_model_path}'
    scaler_name = '{self.complete_scaler_path}'
    label_encoder_name = '{self.complete_label_encoder_path}'
    dataset_url = '{self.dataset_url}'
    
    result = make_prediction(input_data, model_name, scaler_name, label_encoder_name, dataset_url)
    print(result)        
        '''
        
        with open(f'inference{self.model_name}.py', 'w') as file:
            file.write(interference_code)
        
        yield 'Done'


from ImageModelTrainer import ImageModelTrainer


if __name__ == "__main__":
    dataset_url = "https://idesign-quotation.s3.ap-south-1.amazonaws.com/NO_COMPANYNAME/train.zip"
    hasChanged = True
    task = "image"
    mainType = "DL"
    archType = "default" 
    architecture = {}    
    hyperparameters = {} 

    import json

    def returnArch (data, task, mainType, archType):
        current_task = data[task]

        for i in current_task:
            if  i["type"] == mainType and i["archType"] == archType:
                return i["architecture"], i["hyperparameters"]

    arch_data = {}
    task = "image"
    mainType='DL'
    archType='default'

    with open ('arch.json', 'r') as f:
        arch_data = json.load(f)

    architecture, hyperparameters = returnArch(arch_data, task, mainType, archType)

    trainer = ImageModelTrainer(dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters)
    
    executor = trainer.execute()
    
    for epoch_info in executor:
        if isinstance(epoch_info, dict) and 'epoch' in epoch_info:
            print(f"Epoch {epoch_info['epoch']}: Train Loss: {epoch_info['train_loss']:.4f}, Train Acc: {epoch_info['train_acc']:.4f}, Val Loss: {epoch_info['val_loss']:.4f}, Val Acc: {epoch_info['val_acc']:.4f}")
        else:
            model_obj = epoch_info
            break

    if model_obj:
        print(f"Model Object: {model_obj}")
    else:
        print("Failed to train and upload the model.")


from setuptools import setup, find_packages

setup(
    name='xander-ai',
    version='0.2.4',
    packages=find_packages(),
    install_requires=[
        'pandas', 
        'numpy', 
        'chardet', 
        'tensorflow', 
        'scikit-learn', 
        'openpyxl', 
        'channels', 
        'asgiref', 
        'keras', 
        'keras-tuner', 
        'joblib', 
        'requests', 
        'nltk',
        'matplotlib',
        'xgboost'
    ],
    author='Atulit gaur',
    author_email='atulit23@gmail.com',
    description='Xander - A package to train Classification, Regression, and Image Classification models.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    include_package_data=True,
)


from .AIUnified.RegressionDL import RegressionDL
from .AIUnified.ClassificationDL import ClassificationDL
from .AIUnified.ClassificationDL import ClassificationDL
from .AIUnified.ImageModelTrainer import ImageModelTrainer
from .AIUnified.TextModel import TextModel
import numpy as np
import os
import pandas as pd
import zipfile
import json
import chardet

class Xander:
    def  __init__(self, dataset_path="", model_name="v0", hyperparameters={}, target_col=None, task=""):
        self.hyperparameters = hyperparameters
        self.task = task
        self.dataset_path = dataset_path
        self.architecture = {}
        self.df = None
        self.data_dir = 'extracted_files'
        self.model_name = model_name
        self.target_col = target_col
        self.encoding = ''
        
        self.load()
        
    def remove_whitespace(self, filename):
        return ''.join(filename.split())
    
    def returnArch(self, data, task, mainType, archType):
        current_task = data[task]

        for i in current_task:
            if i["type"] == mainType and i["archType"] == archType:
                return i["architecture"], i["hyperparameters"]
            
    def isText(self, df, columns):
        text = []
        avg_sentence_length_check = []

        for column in columns:
            if df[column].dtype == object:
                text.append(True)
                avg_sentence_length = df[column].dropna().apply(lambda x: len(str(x).split())).mean()
                avg_sentence_length_check.append(avg_sentence_length > 4)
            else:
                text.append(False)
                avg_sentence_length_check.append(False)

        if all(text) and any(avg_sentence_length_check):
            return True
        else:
            return False



    def textToNum(self, finalColumn, x):
        arr = finalColumn.unique()
        indices = np.where(arr == x)[0]
        if indices.size > 0:
            index = indices[0]
            return index
        else:
            return -1
    
    def load(self):
        with open(self.dataset_path, 'rb') as file:
            result = chardet.detect(file.read())
            self.encoding = result['encoding']
        if self.dataset_path.endswith('.csv'):
            self.df = pd.read_csv(self.dataset_path, encoding=self.encoding)
        elif self.dataset_path.endswith('.xlsx'):
            self.df = pd.read_excel(self.dataset_path, encoding=self.encoding)
        elif self.dataset_path.endswith('.zip'):
            self.task ='image'

    def train(self):
        if self.task == 'classification':
            model_trainer = ClassificationDL(
                self.dataset_path, self.architecture, self.hyperparameters, self.model_name, self.target_col
            )
            executor = model_trainer.execute()

            for epoch_info in executor:
                if isinstance(epoch_info, dict) and 'epoch' in epoch_info:
                    print(epoch_info)
                else:
                    print(epoch_info)
                    break
                
        if self.task == 'regression':
            model_trainer = RegressionDL(
                self.dataset_path, self.architecture, self.hyperparameters, self.model_name, self.target_col, self.encoding
            )
            executor = model_trainer.execute()

            for epoch_info in executor:
                if isinstance(epoch_info, dict) and 'epoch' in epoch_info:
                    print(epoch_info)
                else:
                    print(epoch_info)
                    break
                
        if self.task == 'text':
            model_trainer = TextModel(
                self.dataset_path, self.model_name, self.hyperparameters, self.target_col, self.encoding
            )
            for result in model_trainer.execute():
                print(result)

        if self.task == 'image':
            model_trainer = ImageModelTrainer(
                self.dataset_path, self.model_name, self.hyperparameters
            )
            executor = model_trainer.execute()


import pandas as pd
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import AdaBoostRegressor
import xgboost as xgb
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import StandardScaler
import joblib
import requests
import os
import uuid

class RegressionML:
    def __init__(self, dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters):
        self.dataset_url = dataset_url
        self.hasChanged = hasChanged
        self.task = task
        self.mainType = mainType
        self.archType = archType
        self.architecture = architecture
        self.hyperparameters = hyperparameters
        self.api_url = 'https://s3-api-uat.idesign.market/api/upload'
        self.bucket_name = 'idesign-quotation'
        
        self.load_data()
        self.preprocess_data()

    def is_regression(self, y):
        unique_elements = len(set(y))
        return unique_elements / len(y) > 0.1
    
    def load_data(self):
        self.df = pd.read_csv(self.dataset_url)
        self.X = self.df.iloc[:, :-1]
        self.y = self.df.iloc[:, -1]
        
    def preprocess_data(self):
        self.df = self.df.dropna(thresh=len(self.df) * 0.8, axis=1)
        self.df = self.df.dropna()

        numeric_cols = self.df.select_dtypes(include=[np.number]).columns
        self.df = self.df[(np.abs(self.df[numeric_cols] - self.df[numeric_cols].mean()) <= (3 * self.df[numeric_cols].std())).all(axis=1)]

        self.X = self.df.iloc[:, :-1]
        self.y = self.df.iloc[:, -1]

        print(self.X)
        print(self.y)

        self.X = pd.get_dummies(self.X, drop_first=True)

        self.scaler = StandardScaler()
        self.X_scaled = self.scaler.fit_transform(self.X)

        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.X_scaled, self.y, test_size=0.2, random_state=42)

    def create_model(self):
        if self.task == "regression":
            if self.archType == '1':
                print("Using Linear Regression")
                self.model = LinearRegression(**self.hyperparameters)
            elif self.archType == "2":
                print("Using XGBoost Regression")
                self.model = xgb.XGBRegressor(**self.hyperparameters)
            elif self.archType == "3":
                print("Using AdaBoost Regression")
                self.model = AdaBoostRegressor(**self.hyperparameters)
            else:
                raise ValueError(f"Unsupported model type: {self.archType}")

    def train_model(self):
        self.model.fit(self.X_train, self.y_train)
        
    def evaluate_model(self):
        y_pred = self.model.predict(self.X_test)
        self.mse = mean_squared_error(self.y_test, y_pred)
        print(f"MSE: {self.mse}")
        return self.mse
        
    def save_model(self):
        model_filename = f'model{str(uuid.uuid4())}.pkl'
        scaler_filename = f'scaler{str(uuid.uuid4())}.pkl'
        
        joblib.dump(self.model, model_filename)
        joblib.dump(self.scaler, scaler_filename)
        
        self.model_path = model_filename
        self.scaler_path = scaler_filename
        
    def upload_files_to_api(self):
        try:
            files = {
                'bucketName': (None, self.bucket_name),
                'files': open(self.model_path, 'rb')
            }
            response_model = requests.put(self.api_url, files=files)
            response_data_model = response_model.json()
            model_url = response_data_model.get('locations', [])[0] if response_model.status_code == 200 else None
            
            if model_url:
                print(f"Model uploaded successfully. URL: {model_url}")
            else: 
                print(f"Failed to upload model. Error: {response_data_model.get('error')}")
                return None, None
            
            files = {
                'bucketName': (None, self.bucket_name),
                'files': open(self.scaler_path, 'rb')
            }
            response_scaler = requests.put(self.api_url, files=files)
            response_data_scaler = response_scaler.json()
            scaler_url = response_data_scaler.get('locations', [])[0] if response_scaler.status_code == 200 else None
            
            if scaler_url:
                print(f"Scaler uploaded successfully. URL: {scaler_url}")
            else:
                print(f"Failed to upload scaler. Error: {response_data_scaler.get('error')}")
                return model_url, None
            
            return model_url, scaler_url
            
        except requests.exceptions.RequestException as e:
            print(f"An error occurred: {str(e)}")
            return None, None
        
    def execute(self):
        self.create_model()
        self.train_model()
        mse = self.evaluate_model()
        self.save_model()
        model_url, scaler_url = self.upload_files_to_api()
        
        if model_url and scaler_url:
            _id = str(uuid.uuid4())
            model_obj = {
                "modelUrl": model_url,
                "size": os.path.getsize(self.model_path) / (1024 ** 3),
                "id": _id,
                "helpers": [{"scaler": scaler_url}],
                "modelArch": self.archType,
                "hyperparameters": self.hyperparameters,
                "task": self.task_type,
                "datasetUrl": self.dataset_url,
                "epoch_data": [{"epoch": 1, "test_loss": mse}]
            }
            return model_obj
        else:
            return None



import pandas as pd
import numpy as np
from xgboost import XGBRegressor
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, MinMaxScaler
from sklearn.metrics import mean_absolute_error
from scipy.interpolate import make_interp_spline
import matplotlib.pyplot as plt
import joblib
import os
import uuid
import queue
import requests
import re

class Forecaster:
    def __init__(self, dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters, userId):
        self.dataset_url = dataset_url
        self.hasChanged = hasChanged
        self.task = task
        self.mainType = mainType
        self.archType = archType
        self.architecture = architecture
        self.hyperparameters = hyperparameters
        self.userId = userId
        self.api_url = 'https://apiv3.xanderco.in/core/store/'
        self.model_path = f'bestmodel{str(uuid.uuid4())}.pkl'
        self.scaler_path = f'scaler{str(uuid.uuid4())}.pkl'
        self.label_encoder_path = f'label_encoder{str(uuid.uuid4())}.pkl'
        self.directory_path = "models"
        self.complete_model_path = os.path.join(self.directory_path, self.model_path)
        self.complete_scaler_path = os.path.join(self.directory_path, self.scaler_path)
        self.complete_label_encoder_path = os.path.join(self.directory_path, self.label_encoder_path)
        self.date_regex = r'^\d{2,4}[-/]\d{1,2}[-/]\d{1,2}$'
        self.label_encoders = {}
        self.time_step = 60
        
        self.load_data()
        self.preprocess_data()
        
    def is_date_like(self, column):
        return any(re.match(self.date_regex, str(value)) for value in column)
    
    def convert_to_seconds(self, date_column):
        date_column = pd.to_datetime(date_column, format='%d/%m/%y', errors='coerce')
        seconds = (date_column - pd.Timestamp("1970-01-01")) // pd.Timedelta('1s')
        return seconds

    def create_dataset(self, data, time_step=60):
        X, y = [], []
        for i in range(len(data) - time_step - 1):
            X.append(data[i:(i + time_step), 1:])
            y.append(data[i + time_step, len(self.df.columns) - 1])
        return np.array(X), np.array(y)

    def load_data(self):
        self.df = pd.read_csv(self.dataset_url)
        self.df = self.df.dropna()

        date_like_columns = [col for col in self.df.columns if self.is_date_like(self.df[col])]
        for column in date_like_columns:
            self.df[column] = self.convert_to_seconds(self.df[column])

        for column in self.df.select_dtypes('object').columns:
            le = LabelEncoder()
            self.df[column] = le.fit_transform(self.df[column])
            self.label_encoders[column] = le

    def preprocess_data(self):
        self.scaler = MinMaxScaler(feature_range=(0, 1))
        self.scaled_data = self.scaler.fit_transform(self.df)
        
        self.X, self.y = self.create_dataset(self.scaled_data, self.time_step)
        self.X = self.X.reshape(self.X.shape[0], -1)
        
        train_size = int(len(self.X) * 0.8)
        self.X_train = self.X[:train_size]
        self.X_test = self.X[train_size:]
        self.y_train = self.y[:train_size]
        self.y_test = self.y[train_size:]

    def create_future_prediction(self, model, last_sequence, steps=60):
        future_predictions = []
        current_sequence = last_sequence.copy()
        
        for _ in range(steps):
            next_pred = model.predict(current_sequence.reshape(1, -1))[0]
            future_predictions.append(next_pred)
            current_sequence = np.roll(current_sequence, -1)
            current_sequence[-1] = next_pred
        
        return future_predictions

    def create_smooth_curve(self, x, y, num_points=300):
        x_smooth = np.linspace(x.min(), x.max(), num_points)
        spl = make_interp_spline(x, y, k=3)
        y_smooth = spl(x_smooth)
        return x_smooth, y_smooth

    def plot_training_results(self):
        time_steps = np.arange(len(self.y_test))
        x_actual_smooth, y_actual_smooth = self.create_smooth_curve(time_steps, self.y_test)
        x_pred_smooth, y_pred_smooth = self.create_smooth_curve(time_steps, self.predictions)

        plt.figure(figsize=(12, 6))
        plt.grid(True, linestyle='-', alpha=0.2)
        plt.plot(x_actual_smooth, y_actual_smooth, '-', color='#4472C4', label='Actual', linewidth=2)
        plt.plot(x_pred_smooth, y_pred_smooth, '-', color='#ED7D31', label='Predicted', linewidth=2)

        plt.title('Growth Curve Forecast', pad=20, fontsize=12)
        plt.xlabel('Time Steps', labelpad=10)
        plt.ylabel('Values', labelpad=10)
        plt.legend()

        ax = plt.gca()
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_alpha(0.5)
        ax.spines['bottom'].set_alpha(0.5)

        plt.tight_layout()
        plt.show()

    def plot_future_predictions(self):
        future_steps = np.arange(len(self.y_test), len(self.y_test) + 60)
        x_future_smooth, y_future_smooth = self.create_smooth_curve(future_steps, np.array(self.future_predictions))

        plt.figure(figsize=(12, 6))
        plt.plot(x_future_smooth, y_future_smooth, '-', color='#A5A5A5', label='Next 60 Steps', linewidth=2)

        plt.title('Future 60-Step Forecast', pad=20, fontsize=12)
        plt.xlabel('Time Steps', labelpad=10)
        plt.ylabel('Values', labelpad=10)
        plt.legend()

        plt.grid(True, linestyle='-', alpha=0.2)
        ax = plt.gca()
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_alpha(0.5)
        ax.spines['bottom'].set_alpha(0.5)

        plt.tight_layout()
        plt.show()

    def create_model(self):
        self.model = XGBRegressor()

    def train_model(self):
        self.epoch_data = []
        self.current_val_acc = 0
        self.epoch_info_queue = queue.Queue()

        self.model.fit(self.X_train, self.y_train)
        self.predictions = self.model.predict(self.X_test)
        self.predictions_train = self.model.predict(self.X_train)
        
        last_sequence = self.X_test[-1]
        self.future_predictions = self.create_future_prediction(self.model, last_sequence)
        
        mae_test = mean_absolute_error(self.y_test, self.predictions)
        mae_train = mean_absolute_error(self.y_train, self.predictions_train)
        print(f"Mean Absolute Error: {mae_test}")
        print(f"Mean Absolute Error: {mae_train}")

        epoch_info = {
            "epoch": 1,
            "train_loss": mae_train,
            "test_loss": mae_test,
        }
        
        self.epoch_data.append(epoch_info)
        
        # self.plot_training_results()
        # self.plot_future_predictions()

    def save_model(self):
        if not os.path.exists(self.directory_path):
            os.makedirs(self.directory_path)
        
        joblib.dump(self.model, self.complete_model_path)
        joblib.dump(self.scaler, self.complete_scaler_path)
        joblib.dump(self.label_encoders, self.complete_label_encoder_path)

    def upload_files_to_api(self):
        print(os.path.getsize(self.complete_model_path) / (1024 ** 3))
        try:
            file = {
                'file': open(self.complete_model_path, 'rb')
            }
            response_model = requests.post(self.api_url, files=file)
            response_data_model = response_model.json()
            model_url = response_data_model.get('file_url')

            if model_url:
                print(f"Model uploaded successfully. URL: {model_url}")
            else:
                print(
                    f"Failed to upload model. Error: {response_data_model.get('error')}")

            file = {
                'file': open(self.complete_scaler_path, 'rb')
            }

            response_scaler = requests.post(self.api_url, files=file)
            response_data_scaler = response_scaler.json()
            scaler_url = response_data_scaler.get('file_url')

            if scaler_url:
                print(f"Scaler uploaded successfully. URL: {scaler_url}")
            else:
                print(
                    f"Failed to upload scaler. Error: {response_data_scaler.get('error')}")
                return model_url, None
            
            file = {
                'file': open(self.complete_label_encoder_path, 'rb')
            }

            response_label = requests.post(self.api_url, files=file)
            response_data_label = response_label.json()
            label_url = response_data_label.get('file_url')

            if label_url:
                print(f"label uploaded successfully. URL: {label_url}")
            else:
                print(
                    f"Failed to upload label. Error: {response_data_label.get('error')}")
            
            print(model_url, scaler_url, label_url)
            return model_url, scaler_url, label_url

        except requests.exceptions.RequestException as e:
            print(f"An error occurred: {str(e)}")
            return None, None

    def execute(self):
        self.create_model()
        self.train_model()
        self.save_model()
        
        model_url, scaler_url, label_url = self.upload_files_to_api()
        
        _id = str(uuid.uuid4())
        
        model_obj = {
            "modelUrl": model_url if model_url and scaler_url else "",
            "size": os.path.getsize(self.complete_model_path) / (1024 ** 3) if model_url and scaler_url else 0,
            "id": _id if model_url and scaler_url else "",
            "helpers": [{"scaler": scaler_url}, {"label_encoders": label_url}] if model_url and scaler_url else [],
            "modelArch": self.architecture,
            "hyperparameters": self.hyperparameters,
            "epoch_data": self.epoch_data,
            "task": self.task,
            "datasetUrl": self.dataset_url
        }
        print(model_obj)
        return model_obj if model_url and scaler_url else None

# forecaster = Forecaster("https://xanderco-storage.s3-accelerate.amazonaws.com/train5109f33b-7226-4142-ab3c-2c7cc7a57d68.csv", False, "", "", "", {}, {}, "84")

# model = forecaster.execute()
# print(model)


import requests
import json
import re
import nltk
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from sklearn.neighbors import NearestNeighbors
from sentence_transformers import SentenceTransformer, util
import shutil
import torch
import zipfile
import uuid
import os

class Chatbot:
    def __init__(self, dataset_url, hasChanged, task, mainType, archType, architecture, hyperparameters, userId):
        self.dataset_url = dataset_url
        self.hasChanged = hasChanged
        self.task = task
        self.mainType = mainType
        self.archType = archType
        self.architecture = architecture
        self.hyperparameters = hyperparameters
        self.api_url = 'https://apiv3.xanderco.in/core/store/'
        self.model = SentenceTransformer('all-MiniLM-L6-v2')
        self.stop_words = set(stopwords.words('english'))
        self.json_url = dataset_url
        self.qa_data = self.fetch_json_data(self.json_url)
        self.questions = [item['question'] for item in self.qa_data]
        self.answers = [item['answer'] for item in self.qa_data]
        self.directory_path = "models"
        self.que_path = f"question_embeddings{str(uuid.uuid4())}.pt"
        self.ans_path = f'answer_embeddings{str(uuid.uuid4())}.pt'
        self.que_complete_path = os.path.join(self.directory_path, self.que_path)
        self.ans_complete_path = os.path.join(self.directory_path, self.ans_path)
        self.userId = userId

    def fetch_json_data(self, url):
        try:
            response = requests.get(url)
            response.raise_for_status()
            data = response.json()
            return data
        except requests.exceptions.RequestException as e:
            print(f"Error fetching JSON data: {str(e)}")
            return None

    def preprocess_text(self, text):
        text = text.lower()
        text = re.sub(r'[^a-zA-Z0-9\s]', ' ', text)
        text = re.sub(r'\s+', ' ', text).strip()
        tokens = word_tokenize(text)
        tokens = [word for word in tokens if word not in self.stop_words]
        return ' '.join(tokens)

    def encode_embeddings(self):
        processed_questions = [self.preprocess_text(
            question) for question in self.questions]
        processed_answers = [self.preprocess_text(
            answer) for answer in self.answers]

        question_embeddings = self.model.encode(
            processed_questions, convert_to_tensor=True)
        answer_embeddings = self.model.encode(
            processed_answers, convert_to_tensor=True)

        return question_embeddings, answer_embeddings

    def upload_files_to_s3(self):
        uploaded_urls = {}

        files_to_upload = [self.que_complete_path, self.ans_complete_path]

        for file_path in files_to_upload:
            file = {
                'file': open(file_path, 'rb')
            }

            try:
                response = requests.post(self.api_url, files=file)
                response_data = response.json()
                print(response_data)
                if response.status_code == 201 or response.status_code == 200:
                    pdf_info = response_data.get('file_url')
                    initial_url = pdf_info
                    uploaded_urls[file_path] = initial_url
                    print(f"File {file_path} uploaded successfully.")
                else:
                    print(
                        f"Failed to upload file {file_path}. Error: {response_data.get('error')}")

            except requests.exceptions.RequestException as e:
                print(f"An error occurred: {str(e)}")

        return uploaded_urls

    def execute(self):
        self.qa_data = self.fetch_json_data(self.dataset_url)

        self.questions = [item['question'] for item in self.qa_data]
        self.answers = [item['answer'] for item in self.qa_data]

        question_embeddings, answer_embeddings = self.encode_embeddings()

        torch.save(question_embeddings,  self.que_complete_path)
        torch.save(answer_embeddings, self.ans_complete_path)

        model_path = "https://xanderco-storage.s3.ap-south-1.amazonaws.com/sentence_transformer_model.zip"

        uploaded_urls = self.upload_files_to_s3()

        _id = str(uuid.uuid4())
        interference_code = f''' 
import requests
import torch
import zipfile
from sentence_transformers import SentenceTransformer
import os
import re 
import nltk
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from sentence_transformers import util
import requests
import json

model_zip_url = '{model_path}'

extract_folder = './sentence_transformer_model'

os.makedirs(extract_folder, exist_ok=True)

print("Downloading model zip file...")
response = requests.get(model_zip_url, stream=True)
zip_file_path = './sentence_transformer_model.zip'

with open(zip_file_path, 'wb') as file:
    for chunk in response.iter_content(chunk_size=1024):
        if chunk:
            file.write(chunk)

print("Unzipping model...")
with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
    zip_ref.extractall(extract_folder)

model_path = os.path.join('sentence_transformer_model')
model = SentenceTransformer(model_path)

question_embeddings_url = '{uploaded_urls.get(self.que_complete_path, "")}'
answer_embeddings_url =  '{uploaded_urls.get(self.ans_complete_path, "")}'

response = requests.get(question_embeddings_url)
with open(question_embeddings_url.split("/")[-1], 'wb') as file:
    file.write(response.content)

response = requests.get(answer_embeddings_url)
with open(answer_embeddings_url.split("/")[-1], 'wb') as file:
    file.write(response.content)

question_embeddings = torch.load(question_embeddings_url.split("/")[-1])
answer_embeddings = torch.load(answer_embeddings_url.split("/")[-1])

nltk.download('punkt')
nltk.download('stopwords')

data_url = '{self.dataset_url}'

response = requests.get(data_url)
qa_data = json.loads(response.text)
questions = [item['question'] for item in qa_data]
answers = [item['answer'] for item in qa_data]

questions = [item['question'] for item in qa_data]
answers = [item['answer'] for item in qa_data]

stop_words = set(stopwords.words('english'))

def preprocess_text(text):
    text = text.lower()
    text = re.sub(r'[^a-zA-Z0-9\s]', ' ', text)  
    text = re.sub(r'\s+', ' ', text).strip()  
    tokens = word_tokenize(text)
    tokens = [word for word in tokens if word not in stop_words]
    return ' '.join(tokens)

def get_answer(question):
    processed_question = preprocess_text(question)
    question_embedding = model.encode(processed_question, convert_to_tensor=True)
    
    similarities = util.pytorch_cos_sim(question_embedding, question_embeddings)[0]
    top_results = similarities.topk(k=5)
    print(top_results)
    similarity, index = similarities.max(), similarities.argmax()
    similarity_percentage = similarity.item() * 100
    
    if similarity_percentage > 45:
        return answers[index], similarity_percentage
    else: 
        return "Sorry, I didn't understand that!", similarity_percentage

user_question = "What is munafa"
answer, similarity_percentage = get_answer(user_question)

print(f"Answer: {{answer}}")
print(f"Similarity Percentage: {{similarity_percentage:.2f}}%")
        '''

        api_code_python = f'''
import requests
import json

url = "https://apiv3.xanderco.in/core/interference/" 

data = {{
    "data": "Your input text",
    "modelId": '{_id}',
    "userId": '{self.userId}',
}}

try:
    response = requests.post(url, json=data)

    if response.status_code == 200:
        # Print the response JSON
        print("Response:")
        print(json.dumps(response.json(), indent=2))
    else:
        print(f"Error: {{response.status_code}}")
        print(response.text)
except requests.exceptions.RequestException as e:
    print(f"An error occurred: {{e}}")
'''

        api_code_js = f'''
const url = "https://apiv3.xanderco.in/core/interference/";

const data = {{
    data: "Your text here",
    modelId: '{_id}',
    userId: '{self.userId}',
}};

const headers = {{
    'Content-Type': 'application/json',
}};

fetch(url, {{
    method: 'POST',
    headers: headers,
    body: JSON.stringify(data)
}})
.then(response => response.json().then(data => {{
    if (response.ok) {{
        console.log("Response:");
        console.log(JSON.stringify(data, null, 2));
    }} else {{
        console.error(`Error: ${{response.status}}`);
        console.error(data);
    }}
}}))
.catch(error => {{
    console.error(`An error occurred: ${{error}}`);
}});
'''
        
        model_obj = {
            "modelUrl": model_path,
            "helpers": [{"question_embeddings": uploaded_urls.get(self.que_complete_path, "")}, {"answer_embeddings": uploaded_urls.get(self.ans_complete_path, "")}],
            "id": _id,
            "architecture": "Sentence Transformers",
            "hyperparameters": {},
            "size": os.path.getsize(self.que_complete_path) / (1024 ** 3) + os.path.getsize(self.ans_complete_path) / (1024 ** 3),
            "task": self.task,
            "interferenceCode": interference_code,
            "datasetUrl": self.dataset_url,
                "codes": [
                    {"python": api_code_python},
                    {"javascript": api_code_js}
                ]
        }

        # os.remove(self.que_path)
        # os.remove(self.ans_path)
        return model_obj
